(*************************************************************************)
(*                                                                       *)
(*                                OCaPIC                                 *)
(*                                                                       *)
(*            Benoit Vaugon, Université Pierre et Marie Curie            *)
(*                                                                       *)
(*    Ce fichier est distribué sous les termes de la licence CeCILL-B    *)
(*    décrite dans le fichier ../../LICENCE.                             *)
(*                                                                       *)
(*************************************************************************)

exception Exn of string

let send_looping_time = 20000;;
let verbose = false;;

type 'a channel = {
  input : Pervasives.in_channel;
  output : Pervasives.out_channel;
}

let open_tty filename = {
  input = open_in_bin filename;
  output = open_out_bin filename;
}

let open_prog prog =
  let ((i, o) as p) = Unix.open_process prog in
    at_exit (fun () -> ignore (Unix.close_process p));
    { input = i ; output = o }
;;

type flag =
  | New | Ok | Adr | End | Data | Error
  | Resend | Already | Oom | Invalid of int

let string_of_flag f =
  match f with
    | New -> "New"
    | Ok -> "Ok"
    | Adr -> "Adr"
    | End -> "End"
    | Data -> "Data"
    | Error -> "Error"
    | Resend -> "Resend"
    | Already -> "Already"
    | Oom -> "Oom"
    | Invalid d -> "Invalid (" ^ (string_of_int d) ^ ")"
;;

let crc_init  = 0x8CB1;;
let flag_base = 0xA6;;

let new_flag     = flag_base lxor 0b00000000;;
let ok_flag      = flag_base lxor 0b11111111;;
let adr_flag     = flag_base lxor 0b10101010;;
let end_flag     = flag_base lxor 0b01010101;;
let data_flag    = flag_base lxor 0b00111100;;
let error_flag   = flag_base lxor 0b11001100;;
let resend_flag  = flag_base lxor 0b00110011;;
let already_flag = flag_base lxor 0b10011001;;
let oom_flag     = flag_base lxor 0b01100110;;

let esc_flag  = 0xFF;;
let res1_flag = 0x03;;
let res2_flag = 0x1A;;
let res3_flag = 0x1C;;

let flag_of_byte b =
  if b = new_flag then New
  else if b = ok_flag then Ok
  else if b = adr_flag then Adr
  else if b = end_flag then End
  else if b = data_flag then Data
  else if b = error_flag then Error
  else if b = resend_flag then Resend
  else if b = already_flag then Already
  else if b = oom_flag then Oom
  else Invalid b
;;

let byte_of_flag f =
  match f with
    | New -> new_flag
    | Ok -> ok_flag
    | Adr -> adr_flag
    | End -> end_flag
    | Data -> data_flag
    | Error -> error_flag
    | Resend -> resend_flag
    | Already -> already_flag
    | Oom -> oom_flag
    | Invalid _ -> failwith "Internal error"
;;

let crc tab =
  let size = Array.length tab in
  let rec f acc i =
    if i = size then (acc land 0xFF, acc lsr 8)
    else
      let d = tab.(i) in
      let rec g acc mask n =
	if n = 0 then acc
	else
	  let acc0 =
	    if d land mask = 0 then (acc lsl 1)
	    else ((acc lsl 1) lor 1)
	  in
	  let acc1 =
	    if acc0 land 0x10000 = 0 then acc0
	    else acc0 lxor 0x18005
	  in
	    g acc1 (mask lsr 1) (pred n)
      in
	f (g acc (0b10000000) 8) (succ i)
  in
    f crc_init 0
;;

let normalize value =
  let obj = Obj.repr value in
    if Obj.is_int obj || Obj.size obj <> 0 then obj
    else
      if Obj.tag obj = 0 then Obj.repr ()
      else raise (Exn "normalize: invalid block (size = 0 && tag <> 0)")
;;

let serialize =
  let error msg = raise (Exn ("serialize: " ^ msg)) in
  let buflen = 2000 in
  let buf = Array.make buflen 0 in
fun obj ->
  let adr = ref 0 in
  let stored = ref [] in
  let store_byte b =
    let a = !adr in
      if a >= buflen then error "too big value";
      buf.(a) <- b;
      incr adr;
  in
  let curr_adr () = !adr in
  let search_obj obj =
    try Some (List.assq obj !stored)
    with Not_found -> None
  in
  let add_obj obj adr = stored := (obj, adr) :: !stored in
  let store_float f =
    let (x, ex) =
      let (_, ex) as r = frexp f in
	if ex < -128 then frexp 0.0 else r
    in
    let rec d i a r =
      if i < 0 then
	r
      else if a < 0.5 then
	d (pred i) (a *. 2.) r
      else
	d (pred i) ((a -. 0.5) *. 2.) (r lor (1 lsl i))
    in
    let s = if f < 0. then 1 else 0 in
    let mx = d 24 (abs_float x) 0 land 0xFFFFFF in
    let m = (mx lsr 1) + (mx land 1) in
      store_byte (m land 0xFF);
      store_byte ((m lsr 8) land 0xFF);
      store_byte (((m lsr 16) land 0x7F) lor (s lsl 7));
      store_byte ((ex + 126) land 0xFF);
  in
  let rec f uno =
    add_obj uno (curr_adr () + 2);
    let o = normalize uno in
    let tag = Obj.tag o in
    let size = Obj.size o in
      if tag = Obj.lazy_tag then error "lazy block"
      else if tag = Obj.closure_tag then error "closure"
      else if tag = Obj.object_tag then error "object"
      else if tag = Obj.infix_tag then error "closure"
      else if tag = Obj.forward_tag then f (Obj.field o 0)
      else if tag = Obj.abstract_tag then error "abstract block"
      else if tag = Obj.string_tag then (
	let str = Obj.obj o in
	let len = String.length str in
	let sz = len / 2 + 1 in
	  if len > 509 then error "too big string (size > 509)";
	  store_byte tag;
	  store_byte sz;
	  let my_adr = curr_adr () in
	    for i = 0 to pred len do
	      store_byte (int_of_char str.[i]);
	    done;
	    store_byte 0;
	    if len mod 2 = 0 then store_byte 1;
	    my_adr
      ) else if tag = Obj.double_tag then (
	let (fl : float) = Obj.obj o in
	  store_byte tag;
	  store_byte 2;
	  let my_adr = curr_adr () in
	    store_float fl;
	    my_adr
      ) else if tag = Obj.double_array_tag then (
	if size > 127 then error "too big float array (size > 127)";
	store_byte tag;
	store_byte (2 * size);
	let my_adr = curr_adr () in
	  for i = 0 to size do
	    let fi = Obj.double_field o i in
	      store_float fi;
	  done;
	  my_adr
      ) else if tag = Obj.custom_tag then error "custom block"
      else if tag = Obj.out_of_heap_tag then error "out of heap block"
      else if tag = Obj.unaligned_tag then error "unaligned adress"
      else if tag > 255 then error "invalid block tag (> 255)"
      else if size = 0 then failwith "Internal error"
      else (
	if size > 255 then error "too big block (size > 255)";
	store_byte tag;
	store_byte size;
	let my_adr = curr_adr () in
	  adr := !adr + size * 2;
	  if !adr >= buflen then error "too big value";
	  for i = 0 to pred size do
	    let unfi = Obj.field o i in
	    let fi = normalize unfi in
	      if Obj.is_int fi then (
		let (n : int) = Obj.obj fi in
		  if n < -16384 || n > 16383 then
		    error (Printf.sprintf
			     "integer %d out of bounds [ -16384 ; 16383 ]" n);
		  buf.(my_adr + 2 * i) <- ((n lsl 1) lor 1) land 0xFF;
		  buf.(my_adr + 2 * i + 1) <- (n lsr 7) land 0xFF;
	      ) else (
		let fi_adr =
		  match search_obj unfi with
		    | Some adr -> adr
		    | None -> f unfi
		in
		  buf.(my_adr + 2 * i) <- -(fi_adr land 0xFF);
		  buf.(my_adr + 2 * i + 1) <- fi_adr lsr 8;
	      )
	  done;
	  my_adr
      )
  in
    if f obj <> 2 then failwith "Internal error";
    Array.sub buf 0 !adr
;;

let set_adresses data adr =
  for i = 0 to Array.length data - 1 do
    let d = data.(i) in
      if d < 0 then
	let a = (data.(i+1) lsl 8) - d + adr in
	  data.(i) <- a land 0xFF;
	  data.(i+1) <- a lsr 8;
  done
;;

let flush output =
  flush output;
  for i = 1 to send_looping_time do () done;
;;

let send_byte output b =
  if verbose then Printf.printf "send_byte: 0x%02x\n%!" b;
  if b = res1_flag || b = res2_flag || b = res3_flag || b = esc_flag then (
    output_byte output esc_flag;
    flush output;
    output_byte output (b lxor esc_flag);
    flush output;
  ) else (
    output_byte output b;
    flush output;
  )
;;

let send_flag output f =
  if verbose then Printf.printf "send_flag: %s\n%!" (string_of_flag f);
  send_byte output (byte_of_flag f)
;;

let receive_byte input =
  let b =
    let o = input_byte input in
      if o = esc_flag then (input_byte input) lxor esc_flag
      else o
  in
    if verbose then Printf.printf "receive_byte: 0x%02x\n%!" b;
    b
;;

let receive_flag input =
  let f = flag_of_byte (receive_byte input) in
    if verbose then Printf.printf "receive_flag: %s\n%!" (string_of_flag f);
    f
;;

let send { input = input ; output = output } value =
  let error msg = raise (Exn ("send: " ^ msg)) in
  let obj = normalize value in
  let rec start () =
    send_flag output New;
    match receive_flag input with
      | Ok -> send_header ()
      | _ -> start ()
  and send_header () =
    send_flag output Ok;
    let (accu_or_size, data) =
      if Obj.is_int obj then (
	let (n : int) = Obj.obj obj in
	  if n < -16384 || n > 16383 then
	    error (Printf.sprintf
		     "integer %d out of bounds [ -16384 ; 16383 ]" n);
	  ((n lsl 1) lor 1, [||])
      ) else (
	let tab = serialize obj in
	let size = Array.length tab in
	  (size, tab)
      )
    in
    let b0 = accu_or_size land 0xFF in
    let b1 = accu_or_size lsr 8 in
    let (crc0, crc1) = crc [| b0 ; b1 |] in
      send_byte output b0;
      send_byte output b1;
      send_byte output crc0;
      send_byte output crc1;
      if Obj.is_int obj then check_accu ()
      else check_size data;
  and check_accu () =
    match receive_flag input with
      | End -> ()
      | _ -> resend ()
  and resend () =
    send_flag output Resend;
    match receive_flag input with
      | Already -> ()
      | Ok -> send_header ()
      | _ -> resend ()
  and check_size data =
    match receive_flag input with
      | Oom -> error "too big value"
      | Adr -> receive_adr data
      | _ -> resend ()
  and receive_adr data =
    let buf = "    " in
      buf.[0] <- char_of_int (receive_byte input);
      buf.[1] <- char_of_int (receive_byte input);
      buf.[2] <- char_of_int (receive_byte input);
      buf.[3] <- char_of_int (receive_byte input);
      let b0 = int_of_char buf.[0] in
      let b1 = int_of_char buf.[1] in
      let crc0 = int_of_char buf.[2] in
      let crc1 = int_of_char buf.[3] in
      let (ccrc0, ccrc1) = crc [| b0 ; b1 |] in
	if ccrc0 <> crc0 || ccrc1 <> crc1 then resend ()
	else
	  let adr = b0 lor (b1 lsl 8) in
	    set_adresses data adr;
	    let (crc0, crc1) = crc data in
	      send_flag output Data;
	      Array.iter (send_byte output) data;
	      send_byte output crc0;
	      send_byte output crc1;
	      match receive_flag input with
		| End -> ()
		| _ -> resend_data data
  and resend_data data =
    send_flag output Resend;
    match receive_flag input with
      | Already -> ()
      | Adr -> receive_adr data
      | _ -> resend_data data
  in
    start ()
;;
