(*************************************************************************)
(*                                                                       *)
(*                                OCaPIC                                 *)
(*                                                                       *)
(*            Benoit Vaugon, Université Pierre et Marie Curie            *)
(*                                                                       *)
(*    Ce fichier est distribué sous les termes de la licence CeCILL-B    *)
(*    décrite dans le fichier ../LICENCE.                                *)
(*                                                                       *)
(*************************************************************************)

(** Character type LCD display communication interface. *)

type bus_size = Four | Eight
(** Possible sizes of data connection between the PIC and the LCD display *)

type direction = Left | Right
(** Direction of cursor or display shift. *)
    
type mode = Cursor_left | Cursor_right | Display_left | Display_right
(** Writing mode: cursor or display moving and direction setting. *)
      
type state = On | Off
(** State of the LCD display. *)
      
type cursor = Invisible | Underscore | Block
(** Cursor mode. *)

type lmode = One | Two
(** Line mode. *)
      
type font = F5x8 | F5x10
(** Characters font. *)
      
(** Inforamtions about connection between the PIC and the LCD display. *)
module type S =
sig
  val bus_size : bus_size
    (** Bus size: true -> 8 ; false -> 4. *)

  val e : Pic.bit
    (** Pin of the pic connected to the [e] entry of the LCD display. *)

  val rs : Pic.bit
    (** Pin of the pic connected to the [rs] entry of the LCD display. *)

  val rw : Pic.bit
    (** Pin of the pic connected to the [rw] entry of the LCD display. *)

  val bus : Pic.reg
    (** Port of the pic connected to the bus of the LCD display.
        In case of 4 bit bus, the 4th highest bits of the port are used. *)
end

(** Connects the PIC to the LCD display.
    Takes informations about connection in the module C. *)
module Connect : functor (C : S) ->
sig
  val clear : unit -> unit
    (** Clears the LCD display. *)
    
  val home : unit -> unit
    (** Returns cursor and display to their initial positions. *)
    
  (***)

  val init : unit -> unit
    (** Initialises the LCD display.
        Should be called once on the LCD power on. *)
    
  val config : ?mode:mode -> ?disp:state -> ?cursor:cursor ->
    ?lmode:lmode -> ?font:font -> unit -> unit
    (** Configures the LCD display.
        Should be called after the LCD initialisation. *)
    
  (***)

  val current_position : unit -> int * int
    (** Get the current position of the cursor as (line, column). *)

  val moveto : int -> int -> unit
    (** moveto line column  moves the cursor at (line, column). *)
    
  val shift_cursor : direction -> unit
    (** Shifts cursor of 1 character. *)
    
  val shift_display : direction -> unit
    (** Shifts display of 1 character. *)
    
  (***)

  val print_char : char -> unit
    (** Write the character on the LCD display. *)

  val print_string : string -> unit
    (** Write the string on the LCD display at the current cursor position. *)

  val print_int : int -> unit
    (** Write one integer on the LCD display, in decimal. *)

  val output : out_channel
    (** Display view as output channel. Can be used as argument of
	Printf.fprintf. *)
    
end

type display = {
  clear : unit -> unit;
  (** Clears the LCD display. *)
  home : unit -> unit;
  (** Returns cursor and display to their initial positions. *)
  (***)
  init : unit -> unit;
  (** Initialises the LCD display.
      Should be called once on the LCD power on. *)
  config : ?mode:mode -> ?disp:state -> ?cursor:cursor ->
		       ?lmode:lmode -> ?font:font -> unit -> unit;
  (** Configures the LCD display.
      Should be called after the LCD initialisation. *)
  (***)
  current_position : unit -> int * int;
  (** Get the current position of the cursor as (line, column). *)
  moveto : int -> int -> unit;
  (** moveto line column  moves the cursor at (line, column). *)
  shift_cursor : direction -> unit;
  (** Shifts cursor of 1 character. *)
  shift_display : direction -> unit;
  (** Shifts display of 1 character. *)
  (***)
  print_char : char -> unit;
  (** Write the character on the LCD display. *)
  print_string : string -> unit;
  (** Write the string on the LCD display at the current cursor position. *)
  print_int : int -> unit;
  (** Write one integer on the LCD display, in decimal. *)
  output : out_channel;
  (** Display view as output channel. Can be used as argument of
      Printf.fprintf. *)
}

val connect : bus_size:bus_size -> e:Pic.bit -> rs:Pic.bit -> rw:Pic.bit ->
  bus:Pic.reg -> display
  (** Create a connection to an Lcd display. Offers the same features as
  Connect but in another style. *)
