(*************************************************************************)
(*                                                                       *)
(*                                OCaPIC                                 *)
(*                                                                       *)
(*            Benoit Vaugon, Université Pierre et Marie Curie            *)
(*                                                                       *)
(*    Ce fichier est distribué sous les termes de la licence CeCILL-B    *)
(*    décrite dans le fichier ../../LICENCE.                             *)
(*                                                                       *)
(*************************************************************************)

let interp = ref (Config.libdir ^ "/interp.asm") in
let runtime = ref (Config.libdir ^ "/runtime.asm") in
let stdlib = ref (Config.libdir ^ "/stdlib.asm") in
let stack_size = ref 1 in
let heap_size = ref 7 in
let output = ref None in
let header = ref None in
let verbose = ref false in
let bytecode = ref None in
let externs = ref [] in
let spec =
  Arg.align [
    ("-o", Arg.String (fun o -> output := Some o),
     "<file.asm> Define output file (default: <bytecode>.asm)");
    ("-header", Arg.String (fun h -> header := Some h),
     "<file.inc> Define pic header file to include");
    ("-interp", Arg.Set_string interp,
     "<file.asm> Define bytecode interpreter");
    ("-runtime", Arg.Set_string runtime,
     "<file.asm> Define assembler runtime");
    ("-stdlib", Arg.Set_string stdlib,
     "<file.asm> Define assembler standard library");
    ("-stack-size", Arg.Set_int stack_size,
     " Define stack size (default: 1 -> 174 levels)");
    ("-heap-size", Arg.Set_int heap_size,
     " Define heap size (default: 7 -> 1792 bytes)");
    ("-verbose", Arg.Set verbose,
     " Verbose mode");
    ("-version", Arg.Unit (fun () -> print_endline Config.version ; exit 0),
     " Print version and exit");
  ] in
let usage =
  Printf.sprintf "Usage: %s <pic> [ OPTIONS ] <bytecode> [ <externs.asm> ... ]"
    Sys.argv.(0)
in
let error msg =
  Printf.printf "Error: %s\n" msg;
  Arg.usage spec usage;
  exit 1;
in
let unknow arg =
  if Filename.check_suffix arg ".asm" then
    externs := arg :: !externs
  else if !bytecode = None then
    bytecode := Some arg
  else
    error (Printf.sprintf "don't know what to do with: `%s'" arg);
in
let pic =
  if Array.length Sys.argv < 2 then error "pic requested";
  if not (List.mem Sys.argv.(1) Config.supported_pics) then
    error (Printf.sprintf "invalid or unsupported pic: `%s'" Sys.argv.(1));
  incr Arg.current;
  Sys.argv.(1)
in
  Arg.parse spec unknow usage;
  let bytecode =
    match !bytecode with
      | None -> error "bytecode file undefined";
      | Some bytecode -> bytecode
  in
  let header =
    match !header with
      | None -> "p" ^ pic ^ ".inc"
      | Some h -> h
  in
  let output =
    match !output with
      | None -> bytecode ^ ".asm"
      | Some o -> o
  in
  let print_include oc = Printf.fprintf oc "        include \"%s\"\n" in
    try
      let ic = open_in_bin bytecode in
      let index = Index.parse ic in
      let prims = Prim.parse ic index in
      let code = Code.parse ic index in
      let data = Data.parse ic index in
      let code = Data.remap_code code in
      let premap = Prim.remap prims code in
	Code.resolve_addr (Prim.length premap) code;
	if !verbose then Data.print stdout data;
	if !verbose then Code.print stdout code;
	Code.check code;
	let oc = open_out_bin output in
	  Printf.fprintf oc "        processor %s\n" pic;
	  output_char oc '\n';
	  print_include oc header;
	  output_char oc '\n';
	  Printf.fprintf oc "STACK_SIZE  EQU   0x%X\n" !stack_size;
	  Printf.fprintf oc "HEAP_SIZE   EQU   0x%X\n" !heap_size;
	  output_char oc '\n';
	  Data.print_useprim oc data;
	  Prim.print_useprim oc premap;
	  output_char oc '\n';
	  print_include oc !interp;
	  print_include oc !runtime;
	  output_char oc '\n';
	  Prim.export oc premap;
	  Code.export oc code;
	  Data.export oc data;
	  List.iter (print_include oc) !externs;
	  print_include oc !stdlib;
	  Printf.fprintf oc "\n        end\n";
	  close_in ic;
	  close_out oc;
    with
      | Code.Exn msg | Data.Exn msg | Index.Exn msg | Instr.Exn msg
      | Prim.Exn msg ->
	  Printf.eprintf "Error: %s\n%!" msg;
	  begin try Sys.remove output with _ -> () end;
	  exit 1;
;;
