/*************************************************************************/
/*                                                                       */
/*                                OCaPIC                                 */
/*                                                                       */
/*            Benoit Vaugon, Université Pierre et Marie Curie            */
/*                                                                       */
/*    Ce fichier est distribué sous les termes de la licence CeCILL-B    */
/*    décrite dans le fichier ../../LICENCE.                             */
/*                                                                       */
/*************************************************************************/

#include <time.h>
#include <stdio.h>
#include <caml/alloc.h>
#include <caml/fail.h>
#include <caml/memory.h>
#include <caml/mlvalues.h>

/* Gc */

CAMLprim value caml_gc_stat(value v);
CAMLprim value caml_gc_full_major(value v);
CAMLprim value caml_gc_minor(value v);

value caml_gc_run(value unit){
  CAMLparam1(unit);
  caml_gc_minor(Val_unit);
  CAMLreturn(Val_unit);
}

value caml_heap_size(value unit){
  return Val_long(0x7 * 256);
}

value caml_stack_size(value unit){
  return Val_long(0xF57 - 0xE00);
}

value caml_gc_compaction(value unit);

value caml_heap_occupation(value unit){
  CAMLparam0();
  CAMLlocal1(stat);
  caml_gc_compaction(unit);
  stat = caml_gc_stat(unit);
  CAMLreturn(Val_long(2 * Long_val(Field(stat, 7))));
}

value caml_stack_occupation(value unit){
  return Val_long(42);
}

value caml_running_number(value unit){
  CAMLparam1(unit);
  CAMLlocal1(stat);
  caml_gc_full_major(Val_unit);
  stat = caml_gc_stat(Val_unit);
  CAMLreturn(Field(stat, 3)); /* minor_collection */
}

/* Random */

value caml_random_bits(value unit){
  return Val_long(random() % 16384);
}

value caml_random_bool(value unit){
  return Val_long(random() % 2 != 0);
}

value caml_random_round(value unit){
  srandom(time(NULL));
  return Val_unit;
}

/* Interrupt */

value caml_set_interruption_handler(value unit){
  return Val_unit;
}

/* Tool */

value caml_raise_ios_failure(value unit){
  caml_failwith("int_of_string");
}

/* Sleep */

value caml_sleep_millis(value time){
  int nb = Long_val(time);
  if(nb <= 0) return Val_unit;
  struct timespec t;
  t.tv_sec  = nb / 1000;
  t.tv_nsec = (nb % 1000) * 1000000;
  nanosleep(&t, NULL);
  return Val_unit;
}

/* Serial */

#define SERIAL_BUFFER_SIZE 2000

#define SERIAL_CRC_INIT  0x8CB1

#define SERIAL_FLAG_BASE    0xA6
#define SERIAL_NEW_FLAG     (SERIAL_FLAG_BASE ^ 0x00)
#define SERIAL_OK_FLAG      (SERIAL_FLAG_BASE ^ 0xFF)
#define SERIAL_ADR_FLAG     (SERIAL_FLAG_BASE ^ 0xAA)
#define SERIAL_END_FLAG     (SERIAL_FLAG_BASE ^ 0x55)
#define SERIAL_DATA_FLAG    (SERIAL_FLAG_BASE ^ 0x3C)
#define SERIAL_ERROR_FLAG   (SERIAL_FLAG_BASE ^ 0xCC)
#define SERIAL_RESEND_FLAG  (SERIAL_FLAG_BASE ^ 0x33)
#define SERIAL_ALREADY_FLAG (SERIAL_FLAG_BASE ^ 0x99)
#define SERIAL_OOM_FLAG     (SERIAL_FLAG_BASE ^ 0x66)

#define SERIAL_ESC_FLAG  0xFF
#define SERIAL_RES1_FLAG 0x03
#define SERIAL_RES2_FLAG 0x1A
#define SERIAL_RES3_FLAG 0x1C

static unsigned char serial_flag = 0;
static unsigned char serial_crc_byteL;
static unsigned char serial_crc_byteH;

value caml_serial_open_in(value unit){
  return Val_unit;
}

value caml_serial_reception_ready(value unit){
  return Val_false;
}

/* C version of caml_serial_receive of stdlib.asm */

static value caml_serial_receive_wait_new(void);
static value caml_serial_receive_error(void);
static value caml_serial_receive_already_received(void);
static value caml_serial_receive_clear_rcip(void);
static value caml_serial_receive_start(void);
static value caml_serial_receive_heap(int size);
static value caml_serial_receive_data(int size);
static value caml_serial_receive_data_loop(int size);
static value caml_serial_receive_end(value ret);
static value caml_serial_receive_wait_resend(void);
static value caml_serial_receive_oom(void);
static value caml_serial_receive_data_error(int size);
static value caml_serial_receive_data_flag_error(int size);
static unsigned char caml_serial_receive_byte(void);
static void caml_serial_send_byte(unsigned char b);
static void caml_compute_crc_init(void);
static void caml_compute_crc_step(unsigned char b);

value caml_serial_receive(value unit){
  if(serial_flag){
    return caml_serial_receive_clear_rcip();
  }else{
    return caml_serial_receive_wait_new();
  }
}

static value caml_serial_receive_wait_new(void){
  switch(caml_serial_receive_byte()){
  case SERIAL_NEW_FLAG:
    return caml_serial_receive_start();
  case SERIAL_RESEND_FLAG:
    return caml_serial_receive_already_received();
  default:
    return caml_serial_receive_error();
  }
}

static value caml_serial_receive_error(void){
  caml_serial_send_byte(SERIAL_ERROR_FLAG);
  return caml_serial_receive_wait_new();
}

static value caml_serial_receive_already_received(void){
  caml_serial_send_byte(SERIAL_ALREADY_FLAG);
  return caml_serial_receive_wait_new();
}

static value caml_serial_receive_clear_rcip(void){
  serial_flag = 0;
  return caml_serial_receive_start();
}

static value caml_serial_receive_start(void){
  caml_serial_send_byte(SERIAL_OK_FLAG);
  switch(caml_serial_receive_byte()){
  case SERIAL_NEW_FLAG:
    return caml_serial_receive_start();
  case SERIAL_OK_FLAG: {
    unsigned char byteL, byteH, crcL, crcH;
    caml_compute_crc_init();
    byteL = caml_serial_receive_byte();
    caml_compute_crc_step(byteL);
    byteH = caml_serial_receive_byte();
    caml_compute_crc_step(byteH);
    crcL = caml_serial_receive_byte();
    crcH = caml_serial_receive_byte();
    if(crcL != serial_crc_byteL || crcH != serial_crc_byteH){
      return caml_serial_receive_wait_resend();
    }else{
      int accu_or_size = (((int) byteH) << 8) | (int) byteL;
      if(byteL & 1){
	return caml_serial_receive_end(Val_long(accu_or_size >> 1));
      }else{
	return caml_serial_receive_heap(accu_or_size);
      }
    }
  }
  default:
    return caml_serial_receive_error();
  }
}

static value caml_serial_receive_heap(int size){
  unsigned char fsr1L = 0, fsr1H = 0;
  caml_serial_send_byte(SERIAL_ADR_FLAG);
  caml_compute_crc_init();
  caml_serial_send_byte(fsr1L);
  caml_compute_crc_step(fsr1L);
  caml_serial_send_byte(fsr1H);
  caml_compute_crc_step(fsr1H);
  caml_serial_send_byte(serial_crc_byteL);
  caml_serial_send_byte(serial_crc_byteH);
  if(caml_serial_receive_byte() != SERIAL_DATA_FLAG){
    return caml_serial_receive_data_flag_error(size);
  }else{
    return caml_serial_receive_data(size);
  }
}

static value caml_serial_receive_data(int size){
  caml_compute_crc_init();
  return caml_serial_receive_data_loop(size);
}

static value caml_serial_receive_data_loop(int size){
  CAMLparam0();
  CAMLlocal3(block, str, res);
  CAMLlocalN(blocks, SERIAL_BUFFER_SIZE);
  unsigned char buffer[SERIAL_BUFFER_SIZE], crcL, crcH;
  int i, j;
  for(i = 0 ; i < SERIAL_BUFFER_SIZE ; i ++){
    blocks[i] = Val_unit;
  }
  for(i = 0 ; i < size ; i ++){
    unsigned char b = caml_serial_receive_byte();
    buffer[i] = b;
    caml_compute_crc_step(b);
  }
  crcL = caml_serial_receive_byte();
  crcH = caml_serial_receive_byte();
  if(crcL != serial_crc_byteL || crcH != serial_crc_byteH){
    res = caml_serial_receive_data_error(size);
  }else{
    i = 0;
    while(i < size){
      tag_t tag = buffer[i ++];
      mlsize_t block_sz = buffer[i ++];
      switch(tag){
      case Abstract_tag:
	fprintf(stderr, "Error: deserialisation of an abstract block\n");
	exit(1);
      case String_tag: {
	int str_sz = 2*block_sz - 1 - buffer[i - 1 + 2*block_sz];
	str = caml_alloc_string(str_sz);
	for(j = 0 ; j < str_sz ; j ++){
	  String_val(str)[j] = buffer[i + j];
	}
	blocks[i] = str;
	break;
      }
      case Double_tag:
	fprintf(stderr, "Error: deserialisation of a double\n");
	exit(1);
      case Double_array_tag:
	fprintf(stderr, "Error: deserialisation of a double array block\n");
	exit(1);
      case Custom_tag:
	fprintf(stderr, "Error: deserialisation of a custom block\n");
	exit(1);
      default: {
	block = caml_alloc(block_sz, tag);
	blocks[i] = block;
	break;
      }
      }
      i += 2 * block_sz;
    }
    for(i = 0 ; i < size ; i ++){
      block = blocks[i];
      if(block != Val_unit){
	mlsize_t block_sz = buffer[i - 1];
	tag_t tag = buffer[i - 2];
	if(tag < Abstract_tag){
	  for(j = 0 ; j < block_sz ; j ++){
	    int v = (((int) buffer[i + 2*j + 1]) << 8) | buffer[i + 2*j];
	    if(v & 1){
	      Field(block, j) = Val_long(v >> 1);
	    }else{
	      Field(block, j) = blocks[v];
	    }
	  }
	}
      }
    }
    if(0) caml_serial_receive_oom();
    res = caml_serial_receive_end(blocks[2]);
  }
  CAMLreturn (res);
}

static value caml_serial_receive_end(value ret){
  caml_serial_send_byte(SERIAL_END_FLAG);
  return ret;
}

static value caml_serial_receive_wait_resend(void){
  caml_serial_send_byte(SERIAL_ERROR_FLAG);
  if(caml_serial_receive_byte() == SERIAL_RESEND_FLAG){
    return caml_serial_receive_start();
  }else{
    return caml_serial_receive_wait_resend();
  }
}

static value caml_serial_receive_oom(void){
  caml_serial_send_byte(SERIAL_OOM_FLAG);
  return caml_serial_receive_wait_new();
}

static value caml_serial_receive_data_error(int size){
  return caml_serial_receive_data_flag_error(size);
}

static value caml_serial_receive_data_flag_error(int size){
  caml_serial_send_byte(SERIAL_ERROR_FLAG);
  if(caml_serial_receive_byte() == SERIAL_RESEND_FLAG){
    return caml_serial_receive_heap(size);
  }else{
    return caml_serial_receive_data_error(size);
  }
}

/* routines */

static unsigned char caml_serial_receive_byte(void){
  unsigned char b = getc(stdin);
  if(b == SERIAL_ESC_FLAG){
    return getc(stdin) ^ SERIAL_ESC_FLAG;
  }else{
    return b;
  }
}

static void caml_serial_send_byte(unsigned char b){
  switch(b){
  case SERIAL_RES1_FLAG:
  case SERIAL_RES2_FLAG:
  case SERIAL_RES3_FLAG:
  case SERIAL_ESC_FLAG:
    putc(SERIAL_ESC_FLAG, stdout);
    putc(b ^ SERIAL_ESC_FLAG, stdout);
    fflush(stdout);
    break;
  default:
    putc(b, stdout);
    fflush(stdout);
  }
}

static void caml_compute_crc_init(void){
  serial_crc_byteL = SERIAL_CRC_INIT & 0xFF;
  serial_crc_byteH = SERIAL_CRC_INIT >> 8;
}

static void caml_compute_crc_step(unsigned char b){
  int i;
  for(i = 0 ; i < 8 ; i ++){
    int MSBb = b & 0x80;
    int MSB1 = serial_crc_byteL & 0x80;
    int MSB2 = serial_crc_byteH & 0x80;
    b <<= 1;
    serial_crc_byteL <<= 1;
    serial_crc_byteH <<= 1;
    if(MSBb) serial_crc_byteL |= 1;
    if(MSB1) serial_crc_byteH |= 1;
    if(MSB2){
      serial_crc_byteL ^= 0x05;
      serial_crc_byteH ^= 0x80;
    }
  }
}
